# import the required libraries
import numpy as np
import matplotlib.pyplot as plt
from scipy.stats import norm

def estimate_pi(nPoints, seed=None):
    generator = np.random.RandomState(seed=seed)
    xx = generator.uniform(size=nPoints)
    yy = generator.uniform(size=nPoints)
    piEst_4 = np.sum(xx*xx + yy*yy < 1)/nPoints
    piEst = 4*piEst_4
    errorPi = 4*np.sqrt(piEst_4*(1-piEst_4)/(nPoints-1))
    return {'pi':piEst, 'err':errorPi, 'dev':piEst-np.pi, 'pull':(piEst-np.pi)/errorPi}

def plot_dev_pull(nPoints, nBins, nRange, seed=None):
    dp = [estimate_pi(nPoints, seed) for i in range(nRange)]
    dev = [x['dev'] for x in dp]
    pulls = [x['pull'] for x in dp]
    mu, sigma = norm.fit(dev)
    print(' nPoints= ',nPoints,' nBins= ',nBins,' nRange= ',nRange,' dev mu= ',mu,' sigma= ',sigma)
    fig, ax = plt.subplots(2, 1, figsize=(6, 8))
    ax[0].hist(dev, bins=nBins)
    xmin, xmax = ax[0].get_xlim()
    x = np.linspace(xmin, xmax, 100)
    p = nRange*(xmax-xmin)/nBins*norm.pdf(x, mu, sigma)
    ax[0].plot(x, p, 'k', linewidth=2)
    ax[0].set_title(r'$\mathrm{error}\ \mu=%.3f,\ \sigma=%.3f$' %(mu, sigma))
    mu, sigma = norm.fit(pulls)
    print(' nPoints= ',nPoints,' nBins= ',nBins,' nRange= ',nRange,' pulls mu= ',mu,' sigma= ',sigma)
    ax[1].hist(pulls, bins=nBins)
    xmin, xmax = ax[1].get_xlim()
    x = np.linspace(xmin, xmax, 100)
    p = nRange*(xmax-xmin)/nBins*norm.pdf(x, mu, sigma)
    ax[1].plot(x, p, 'k', linewidth=2)
    ax[1].set_title(r'$\mathrm{pull}\ \mu=%.3f,\ \sigma=%.3f$' %(mu, sigma))
    plt.show()

plot_dev_pull(1000, 20, 100000)
#plot_dev_pull(10000, 30, 1000000)
